/**
*
* $Id: Titov V.cpp
* (c) 2005 Sputnik -- erik_h_anderson@hotmail.com
* Based on CVEL Centaur-G code
* Ervins Reinverts -- ervins@gmail.com
* Based on CVEL Fregat code
* (c) 2004 David Henderson -- orbiter@aibs.org.uk
* Licence LGPL -- http://www.gnu.org/licenses/lgpl.txt
*
* Autopilot by rjcroy
*
* You may use this source code in your addon provided that if you
* publish your addon either privately or publicly the end user
* recieves a copy of the source code including any modifications
* you may have made. 
*
*/

#include "Nf104.h"


// 1. vertical lift component (wings and body)
// Borrowed from Mike Norman's Shuttle code

void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*15.0;
	static const double istep = 1.0/step;
	static const int nabsc = 25;
	static const double CL[nabsc] = {0.1, 0.17, 0.2, 0.2, 0.17, 0.1, 0, -0.11, -0.24, -0.38, -0.5, -0.5, -0.02, 0.6355, 0.63, 0.46, 0.28, 0.13, 0.0, -0.16, -0.26, -0.29, -0.24, -0.1, 0.1};
	static const double CM[nabsc] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.0025,0.0012, 0,-0.0012,-0.0007, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	// lift and moment coefficients from -180 to 180 in 15 degree steps.
	// This uses a documented lift slope of 0.0437/deg, everything else is rather ad-hoc

	// change for drag increase at high Mach:
	// Subsonic Equivalents

	double Cl0, Cl1, Cd0, Cd1, Cd2;
	if(M>=5.0) {
	Cl0 = -0.207040;
	Cl1 = 1.675600;
	Cd0 = 0.058540;
	Cd1 = -0.05767;
	Cd2 = 1.3632;

	*cl = Cl0 + Cl1*aoa;
	*cd = Cd0 + Cd1*aoa + Cd2*aoa* aoa;
	}
	aoa += PI;
	int idx = max (0, min (23, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	*cm = (CM[idx] + (CM[idx+1]-CM[idx])*d);
		if(M<1.25) { 
			*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
			*cd = 0.03 + oapiGetInducedDrag (*cl, 2.266, 0.75);
		}
	else if(M<5.0) { 
		*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
		*cd = 0.03 + oapiGetInducedDrag (*cl, 2.266, 0.75);
		}
}

// Lift coefficients for the horizontal tail, which is shadowed by the wing and fuselage
void VLiftCoeff2 (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*15.0;
	static const double istep = 1.0/step;
	static const int nabsc = 25;
	static const double CL[nabsc] = {0.1, 0.17, 0.2, 0.2, 0.17, 0.1, 0, -0.11, -0.24, -0.38, -0.5, -0.5, -0.02, 0.6355, 0.63, 0.46, 0.18, 0.0, 0.0, 0.0, -0.16, -0.29, -0.24, -0.1, 0.1};
	static const double CM[nabsc] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.002, 0.004, 0.0025,0.0012, 0,-0.0012,-0.0007, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	// lift and moment coefficients from -180 to 180 in 15 degree steps.
	// This uses a documented lift slope of 0.0437/deg, everything else is rather ad-hoc

	// change for drag increase at high Mach:
	// Subsonic Equivalents

	double Cl0, Cl1, Cd0, Cd1, Cd2;
	if(M>=5.0) {
	Cl0 = -0.207040;
	Cl1 = 1.675600;
	Cd0 = 0.058540;
	Cd1 = -0.05767;
	Cd2 = 1.3632;

	*cl = Cl0 + Cl1*aoa;
	*cd = Cd0 + Cd1*aoa + Cd2*aoa* aoa;
	}
	aoa += PI;
	int idx = max (0, min (23, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	*cm = (CM[idx] + (CM[idx+1]-CM[idx])*d);
		if(M<1.25) { 
			*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
			*cd = 0.03 + oapiGetInducedDrag (*cl, 2.266, 0.75);
		}
	else if(M<5.0) { 
		*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
		*cd = 0.03 + oapiGetInducedDrag (*cl, 2.266, 0.75);
		}
}

// 2. horizontal lift component (vertical stabilisers and body)

void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {       0,    +0.3,      0,   -0.3,  +0.3,     0,   -0.3,      0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.015 + oapiGetInducedDrag (*cl, 1.5, 0.6) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	
	if (M > 1.0) *cd *= 2.2;
	else *cd *= 0.6;
}


																					

NF104::NF104 (OBJHANDLE hObj, int fmodel)
: COMMON_VESSEL_EX (hObj)
{

	tex = oapiRegisterExhaustTexture("RealExhaust");
	texRCS  = oapiRegisterExhaustTexture ("exhaust_atrcs");

	/* set the propellant handles to NULL so I can check if I need to 
	/* create them in the SetMainStage() function. */
	ph_LH2=NULL;
	ph_Jets=NULL;
	ph_RCS=NULL;
	bAttSet=false;
	bAutoProgrammed=false;
	bControlsMade=false;
	gear_status=DOOR_CLOSED;
	gear_proc=0.0;
	brakes_status=DOOR_CLOSED;
	brakes_proc=0.0;
	flaps_status=DOOR_CLOSED;
	flaps_proc=0.0;
	Vold = _V(0,0,0);
	BrakesOn = false;
	bAbort = false;
	AbortChecked = false;
	atchmass = 0.0;
	dt = 0.0;
	simt = 0.0;
	oldsimt = -0.01;
	displaytime = 0.01;
	d_heading = 270.0;
	last_simt=0.0;
	trimlast = 0.0;
	bEjected = false;
	ignitioncount = -1;


	Dock0 = CreateDock(_V(0,6.16,-3.0),_V(0,1,0),_V(0,0,1));

	MyID=ConnectToOrbiterSoundDLL3(GetHandle());
//		havesound=false;
//	else
//	{
		havesound=true;
		RequestLoadVesselWave3(MyID,PLAYGEAR,"Sound\\NF104\\gear.wav",BOTHVIEW_FADED_FAR);
		RequestLoadVesselWave3(MyID,PLAYJET,"Sound\\NF104\\jet.wav",BOTHVIEW_FADED_FAR);
		RequestLoadVesselWave3(MyID,PLAYBANG,"Sound\\NF104\\stg2sep.wav",BOTHVIEW_FADED_FAR);
		RequestLoadVesselWave3(MyID,PLAYFLAPS,"Sound\\NF104\\flaps.wav",BOTHVIEW_FADED_FAR);

//	}

	anim_gear = 0.0;
	anim_pldoors = 0.0;
	anim_brakes = 0.0;
	anim_nozzles = 0.0;
	anim_trolley = 0.0;

	/* CVE-lite function ! */
	/* IMPORTANT */
	/* This function loops through the pending GDPS requests. Have it run
	/* in the constructor so it can set up things like the InitFuelLevel
	/* (initial fuel level) variable, and load the payloads */
	CheckCentralMessage();   

/*
	PARTICLESTREAMSPEC rps = {
		0, 20, 20, 0, 0.03, 0.5, 100, 3, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_FLAT, 1, 1, PARTICLESTREAMSPEC::ATM_PLIN, 6e7, 12e7
	};
	AddReentryStream (&rps);
	retex = oapiRegisterReentryTexture("reentry");
	SetReentryTexture(retex,6e6,1,1);
*/

	Firstsweep = true;
	AnimationsMade = false;

}


NF104::~NF104()
{

}

void NF104::MakeAnimations()
{

	AnimationsMade = true;

	hNF104					= oapiLoadMeshGlobal ("Nf104");

	
	anim_gear = CreateAnimation (0);
	anim_pldoors = CreateAnimation(0);
	anim_nozzles = CreateAnimation(0);
	anim_elevator = CreateAnimation (0.5);
	anim_elevatortrim = CreateAnimation (0.5);
	anim_raileron = CreateAnimation (0.5);
	anim_laileron = CreateAnimation (0.5);
	anim_rudder = CreateAnimation (0.5);
	anim_brakes = CreateAnimation (0);
	anim_flaps = CreateAnimation (0);
	anim_trolley = CreateAnimation (0);


	static UINT LGearDoorGrp[1] = {8};
	static MGROUP_ROTATE LGearDoor (0, LGearDoorGrp, 1,
		_V(-0.661,-0.580,-1.0), _V(0,0,1), (float)(-100*RAD));
	static UINT LGearGrp[2] = {1, 3};
	static MGROUP_ROTATE LGear (0, LGearGrp, 2,
		_V(-0.516,-0.693,-0.302), _V(1,0,0), (float)( 90*RAD));
	static UINT LGearGrp2[2] = {1, 3};
	static MGROUP_ROTATE LGear2 (0, LGearGrp2, 2,
		_V(-0.516,-0.693,-0.302), _V(0,1,0), (float)(-135*RAD));
	static UINT LGearGrp3[2] = {1, 3};
	static MGROUP_ROTATE LGear3 (0, LGearGrp3, 2,
		_V(-0.516,-0.693,-0.302), _V(0,0,1), (float)(-35*RAD));

	static UINT RGearDoorGrp[1] = {9};
	static MGROUP_ROTATE RGearDoor (0, RGearDoorGrp, 1,
		_V( 0.661,-0.580,-1.0), _V(0,0,1), (float)( 100*RAD));
	static UINT RGearGrp[2] = {4, 5};
	static MGROUP_ROTATE RGear (0, RGearGrp, 2,
		_V( 0.516,-0.693,-0.302), _V(1,0,0), (float)( 90*RAD));
	static UINT RGearGrp2[2] = {4, 5};
	static MGROUP_ROTATE RGear2 (0, RGearGrp2, 2,
		_V( 0.516,-0.693,-0.302), _V(0,1,0), (float)( 135*RAD));
	static UINT RGearGrp3[2] = {4, 5};
	static MGROUP_ROTATE RGear3 (0, RGearGrp3, 2,
		_V( 0.516,-0.693,-0.302), _V(0,0,1), (float)( 35*RAD));

	static UINT LNoseGearDoorGrp[1] = {6};
	static MGROUP_ROTATE LNoseGearDoor (0, LNoseGearDoorGrp, 1,
		_V( 0.161,-0.806, 7.658), _V(0,0.1,0.95), (float)(-90*RAD));
	static UINT RNoseGearDoorGrp[1] = {7};
	static MGROUP_ROTATE RNoseGearDoor (0, RNoseGearDoorGrp, 1,
		_V(-0.161,-0.806, 7.658), _V(0,0.1,0.95), (float)( 90*RAD));
	static UINT NoseGearGrp[2] = {0, 2};
	static MGROUP_ROTATE NoseGear (0, NoseGearGrp, 2,
		_V(0.0,-0.762,7.218), _V(1,0,0), (float)( 90*RAD));

// Known Groups:
 
// 0 Nose Tires
// 1 L Tire
// 2 Nose Gear
// 3 L Gear
// 4 R Gear
// 5 R Tire
// 6 Rt Nose Door
// 7 Lt Nose Door
// 8 Lt Gear Door
// 9 Rt Gear Door
// 10 Fuse
// 11 Ventral
// 12 LStab
// 13 LWing
// 14 Rocket
// 15 RStab
// 16 VStab
// 17 Stab Stub
// 18 L Flap
// 19 L Ail
// 20 Rt Wing
// 21 Rt Flap
// 22 Rt Ail
// 23 Nacelle
// 24 Rudder



	AddAnimationComponent (anim_gear, 0.05, 0.4, &LGearDoor);
	AddAnimationComponent (anim_gear, 0.4, 0.7, &LGear);
	AddAnimationComponent (anim_gear, 0.7, 0.8, &LGear2);
	AddAnimationComponent (anim_gear, 0.8, 0.9, &LGear3);
	AddAnimationComponent (anim_gear, 0.05, 0.4, &RGearDoor);
	AddAnimationComponent (anim_gear, 0.4, 0.7, &RGear);
	AddAnimationComponent (anim_gear, 0.7, 0.8, &RGear2);
	AddAnimationComponent (anim_gear, 0.8, 0.9, &RGear3);
	AddAnimationComponent (anim_gear, 0.4, 0.9, &NoseGear);
	AddAnimationComponent (anim_gear, 0.1, 0.7, &LNoseGearDoor);
	AddAnimationComponent (anim_gear, 0.1, 0.7, &RNoseGearDoor);

	// ***** Elevator animation *****
	static UINT ElevatorGrp[1] = {12};
	static MGROUP_ROTATE Elevator (0, ElevatorGrp, 1,
		_V(0.0, 2.05,-6.0), _V( 1,0,0), (float)( 15*RAD));
	static UINT ElevatorGrp2[1] = {15};
	static MGROUP_ROTATE Elevator2 (0, ElevatorGrp2, 1,
		_V(0.0, 2.05,-6.0), _V( 1,0,0), (float)( 25*RAD));
	AddAnimationComponent (anim_elevator, 0, 1, &Elevator);
	AddAnimationComponent (anim_elevator, 0, 1, &Elevator2);

	// ***** Elevator trim animation *****
	static UINT ElevatorTrimGrp[1] = {12};
	static MGROUP_ROTATE ElevatorTrim (0, ElevatorTrimGrp, 1,
		_V(0.0, 2.05,-6.0), _V( 1,0,0), (float)( 7*RAD));
	AddAnimationComponent (anim_elevatortrim, 0, 1, &ElevatorTrim);
	static UINT ElevatorTrimGrp2[1] = {15};
	static MGROUP_ROTATE ElevatorTrim2 (0, ElevatorTrimGrp2, 1,
		_V(0.0, 2.05,-6.0), _V( 1,0,0), (float)( 7*RAD));
	AddAnimationComponent (anim_elevatortrim, 0, 1, &ElevatorTrim2);

	// ***** Rudder animation *****
	static UINT RudderGrp[1] = {24};
	static MGROUP_ROTATE Rudder (0, RudderGrp, 1,
		_V( 0,1.5,-6.3), _V( 0,0.95,0.1), (float)(-25*RAD));
	AddAnimationComponent (anim_rudder, 0, 1, &Rudder);

	// ***** Aileron animation *****

	static UINT RAileronGrp[1] = {22};
	static MGROUP_ROTATE RAileron (0, RAileronGrp, 1,
		_V( 2.9,-0.644,-1.38), _V( 0.9,-0.15,0), (float)( 30*RAD));
	AddAnimationComponent (anim_raileron, 0, 1, &RAileron);

	static UINT LAileronGrp[1] = {19};
	static MGROUP_ROTATE LAileron (0, LAileronGrp, 1,
		_V(-2.9,-0.644,-1.38), _V(-0.9,-0.15,0), (float)( 30*RAD));
	AddAnimationComponent (anim_laileron, 0, 1, &LAileron);

	
		// ***** Flaps animation *****

	static UINT RFlapGrp[1] = {21};
	static MGROUP_ROTATE RFlap2 (0, RFlapGrp, 1,
		_V( 1.7,-0.45,-1.38),  _V( 0.9,-0.15,0), (float)(-35*RAD));
	static UINT LFlapGrp[1] = {18};
	static MGROUP_ROTATE LFlap2 (0, LFlapGrp, 1,
		_V(-1.7,-0.45,-1.38),  _V(-0.9,-0.15,0), (float)( 35*RAD));


	AddAnimationComponent (anim_flaps, 0, 1, &RFlap2);
	AddAnimationComponent (anim_flaps, 0, 1, &LFlap2);

}


void NF104::MakeAirfoils()
{
	ClearAirfoilDefinitions();
	ClearVariableDragElements();

	CreateVariableDragElement (&brakes_proc, 0.4, _V(0, 0, 0));
	CreateVariableDragElement (&gear_proc, 0.2, _V(0, -0.5, 0));
	CreateVariableDragElement (&flaps_proc, 0.2, _V(0, -0.5, 0));

//                  alignment      position      func      chord  S   AR   
	CreateAirfoil (LIFT_VERTICAL, _V(0,0,+0.3), VLiftCoeff, 2.42, 14.7, 6.8);
	CreateAirfoil (LIFT_VERTICAL, _V(0,0,-5.3), VLiftCoeff2, 1.42, 5.87, 2.4);
	// wing and body lift+drag components

	CreateAirfoil (LIFT_HORIZONTAL, _V(0,0,-4.24), HLiftCoeff, 2.56, 4.42, 6.7);
	// vertical stabiliser and body lift and drag components



}


void NF104::MakeControlSurfaces()
{

//  Aargh!  Can't delete control surfaces
//  Need to make sure we declare once and only once,
//  but that a booster stage alone doesn't get animations for the upper stage!

	if (bControlsMade) return;
	bControlsMade = true;


		CreateControlSurface (AIRCTRL_ELEVATOR,     2.87,  1.4, _V(  0,0.0,-5.3), AIRCTRL_AXIS_XPOS, anim_elevator);
		CreateControlSurface (AIRCTRL_RUDDER,       3.0,  0.7, _V(  0,1.5,-4.5), AIRCTRL_AXIS_YPOS, anim_rudder);
		CreateControlSurface (AIRCTRL_AILERON,      2.0, -0.9, _V(-3.5,0.0,0), AIRCTRL_AXIS_XPOS, anim_raileron);
		CreateControlSurface (AIRCTRL_AILERON,      2.0, -0.9, _V( 3.5,0.0,0), AIRCTRL_AXIS_XNEG, anim_laileron);
		CreateControlSurface (AIRCTRL_FLAP,         3.0, -1.2, _V( 0.0,0.0,0), AIRCTRL_AXIS_XPOS, anim_flaps);
		CreateControlSurface (AIRCTRL_ELEVATORTRIM, 2.0,  1.4, _V(  0,0.0,-5.3), AIRCTRL_AXIS_XPOS, anim_elevatortrim);

}



void NF104::UpdateAnimations()
{	

		// animate landing gear
		if (gear_status >= DOOR_CLOSING) {
			double da = oapiGetSimStep() * GEAR_OPERATING_SPEED;
			if (gear_status == DOOR_CLOSING) {
				if (gear_proc > 0.0)
					gear_proc = max (0.0, gear_proc-da);
				else {
					gear_status = DOOR_CLOSED;
				}
			} else  { // door opening
				if (gear_proc < 1.0)
					gear_proc = min (1.0, gear_proc+da);
				else {
					gear_status = DOOR_OPEN;
				}
			}
			SetAnimation (anim_gear, gear_proc);
		}

		// animate speedbrakes
		if (brakes_status >= DOOR_CLOSING) {
			double da = oapiGetSimStep() * BRAKES_OPERATING_SPEED;
			if (brakes_status == DOOR_CLOSING) {
				if (brakes_proc > 0.0)
					brakes_proc = max (0.0, brakes_proc-da);
				else {
					brakes_status = DOOR_CLOSED;
				}
			} else  { // door opening
				if (brakes_proc < 1.0)
					brakes_proc = min (1.0, brakes_proc+da);
				else {
					brakes_status = DOOR_OPEN;
				}
			}
			SetAnimation (anim_brakes, brakes_proc);
		}

		// animate flaps
		if (flaps_status >= DOOR_CLOSING) {
			double da = oapiGetSimStep() * FLAPS_OPERATING_SPEED;
			if (flaps_status == DOOR_CLOSING) {
				if (flaps_proc > 0.0)
					flaps_proc = max (0.0, flaps_proc-da);
				else {
					flaps_status = DOOR_CLOSED;
				}
			} else  { // door opening
				if (flaps_proc < 1.0)
					flaps_proc = min (1.0, flaps_proc+da);
				else {
					flaps_status = DOOR_OPEN;
				}
			}
			SetControlSurfaceLevel(AIRCTRL_FLAP,flaps_proc);
			SetAnimation (anim_flaps, flaps_proc);
		}

}

void NF104::ActivateLandingGear (DoorStatus action)
{
	gear_status = action;
}

void NF104::RevertLandingGear ()
{
	ActivateLandingGear (gear_status == DOOR_CLOSED || gear_status == DOOR_CLOSING ?
						 DOOR_OPENING : DOOR_CLOSING);
}


void NF104::ActivateBrakes (DoorStatus action)
{
	brakes_status = action;
}

void NF104::RevertBrakes ()
{
	ActivateBrakes (brakes_status == DOOR_CLOSED || brakes_status == DOOR_CLOSING ?
						 DOOR_OPENING : DOOR_CLOSING);
}


void NF104::ActivateFlaps (DoorStatus action)
{
	flaps_status = action;
}

void NF104::RevertFlaps ()
{
	ActivateFlaps (flaps_status == DOOR_CLOSED || flaps_status == DOOR_CLOSING ?
						 DOOR_OPENING : DOOR_CLOSING);
}


void NF104::SetGearParameters (double state)
{

	if ((gear_status == DOOR_OPENING) || (gear_status == DOOR_CLOSING)) {
		double geary = (-0.75*(gear_proc-1.0));
		SetTouchdownPoints (_V(0,-1.68+geary,7.2), _V(-3.6,-1.80+geary,-0.66), _V(3.6,-1.80+geary,-0.66));
	}
	if (gear_proc == 1.0) {
		SetSurfaceFrictionCoeff (0.025, 0.4);
	} else 
	{
		SetSurfaceFrictionCoeff (5, 0.5);		
		if(GroundContact()) 
		{
			//crashed on the ground

			gear_proc = 0.0;
			gear_status = DOOR_OPEN;
		}
		
	}


	double steerforce;
	double steerangle;

	steerforce = 4.0;
	if (GetAirspeed() < 40.0)
		steerforce *=5.0;
	steerforce = steerforce*GetMass();
	if ((GetAirspeed ()-3) < 5.0) steerforce*=((GetAirspeed()-3)/5); // from 1* to 0*
	if (GetAirspeed() < 3.0) steerforce = 0.0;

	steerangle = GetControlSurfaceLevel (AIRCTRL_RUDDER);
	steerforce = steerforce * steerangle;
	if ((GetAltitude() > 4.5) || (gear_proc < 1.0)) steerforce = 0.0;
	AddForce (_V(steerforce, 0, 0), _V(0, 0, 12.0));
	AddForce (_V(-steerforce, 0, 0), _V(0, 0, -12.0));
}


void NF104::TimeStep(  double simt )
{
	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* This function loops through the pending GPDS requests and returns a bool (true/false)
	/* if any GPDS requests have been implemented.
	/* If any have (eg adding a new payload), you will need to re-draw your ship, hence the
	/* call to the (repeatable) SetMainStage(); */
	if( CheckCentralMessage() ) {
		while( CheckCentralMessage() ) ;
			SetMainStage();
	}

	if (!bControlsMade)
		MakeControlSurfaces();

	dt = simt - oldsimt;
	oldsimt = simt;


	if (GetFuelMass() == 0.0)
		if (GetPropellantMass(ph_LH2) == 0.0) {
			SetDefaultPropellantResource (ph_Jets); 
//			SoundOptionOnOff3(MyID,PLAYMAINTHRUST,false);
		}
	if ((GetThrusterGroupLevel(THGROUP_HOVER) > 0.0) && (GetThrusterGroupLevel(THGROUP_HOVER) < 0.90)) {
		SetThrusterGroupLevel(THGROUP_HOVER,1.0);
		SetDefaultPropellantResource (ph_LH2); 
//		SoundOptionOnOff3(MyID,PLAYMAINTHRUST,true);
	}
	if ((GetThrusterGroupLevel(THGROUP_HOVER) > 0.9) && (GetThrusterGroupLevel(THGROUP_HOVER) < 1.0)) {
		SetThrusterGroupLevel(THGROUP_HOVER,0.0);
		SetDefaultPropellantResource (ph_Jets); 
//		SoundOptionOnOff3(MyID,PLAYMAINTHRUST,false);
	}

	if (ignitioncount > 0)
		ignitioncount++;

	if (ignitioncount > 3) {
		ignitioncount = -1;
		AtchV->SetThrusterGroupLevel(THGROUP_MAIN,1.0);
	}

	AirBreathingEngine();


	UpdateAnimations();
	SetGearParameters(gear_proc);

}



void NF104::AirBreathingEngine()
// Thanks to David Hopkins for publishing his AirBreathingEngine source.  
{//=========================================================
	double alt2;
	double eas2;
	double cf1, cf2;
	double ActualThrust;


	alt2 = GetAltitude();
	eas2 = GetAirspeed ();

	alt2 = alt2*3.28*0.7;
	eas2 = eas2*2.236;

	cf1 = 1.67e-6*(eas2*eas2)-8.8789e-4*(eas2)+0.9977;
	cf2 = 3.0e-10*(alt2*alt2)-3.0000e-5*(alt2)+0.9934;

	// Nasty kludge to keep the thrust from increasing dramatically above 50kft
	if (alt2 > 50000)
		cf2 = 0.25*(1-((alt2-50000)*.0001));
	if (cf2 < 0) cf2 = 0;

	if (GetMachNumber() > 2.0)
		cf1 = 10*cf1*(2.2-GetMachNumber());
	if (cf1 < 0) cf1 = 0;

	ActualThrust = sqrt(cf1*cf2)*JET_THRUST;
	if (ActualThrust > JET_THRUST)
		ActualThrust = JET_THRUST;


//	sprintf(oapiDebugString(), "cf1: %.3f, cf2:  %.3f, cf1*cf2:  %.3f ActualThrust:  %.3f", cf1, cf2, cf1*cf2, ActualThrust);

//	double ThrusterActive = GetThrusterGroupLevel(THGROUP_MAIN);
	double ThrusterActive = sqrt(cf1*cf2)*GetThrusterGroupLevel(THGROUP_MAIN);
	if ((GetThrusterGroupLevel(THGROUP_MAIN) > 0.8) && (ThrusterActive > 0.1))		// Afterburner on!
		{
		// we calculate the volume at wich it must be played
		double volume=(110+(145*ThrusterActive));
		double freq=11050+(9000*GetThrusterGroupLevel(THGROUP_MAIN));
		PlayVesselWave3(MyID,PLAYJET,LOOP,volume,freq);								// play the sound
		SoundOptionOnOff3(MyID,PLAYMAINTHRUST,true);								// Rocket noise too
		SetThrusterIsp(th_Jets[0],ISP_JET_VAC);
		trimlast = GetControlSurfaceLevel(AIRCTRL_ELEVATORTRIM);
		}
	else if(ThrusterActive > 0.1)													// Afterburner off
		{
		// we calculate the volume at wich it must be played
		double volume=(110+(145*ThrusterActive));
		double freq=11050+(9000*GetThrusterGroupLevel(THGROUP_MAIN));
		PlayVesselWave3(MyID,PLAYJET,LOOP,volume,freq);								// play the sound
		SoundOptionOnOff3(MyID,PLAYMAINTHRUST,false);								// No rocket noise
		SetThrusterIsp(th_Jets[0],4*ISP_JET_VAC);									// Increased fuel efficiency w/o AB
		ActualThrust *= 0.777;														// 0.8 throttle deflection gives us 0.62*thrust level, so normalize
		trimlast = GetControlSurfaceLevel(AIRCTRL_ELEVATORTRIM);
		}
	else if(IsPlaying3(MyID,PLAYJET))
		{
		// else shut down the sound if it's playing
		StopVesselWave3(MyID,PLAYJET);
		SetThrusterGroupLevel(THGROUP_MAIN,0.0);
		SoundOptionOnOff3(MyID,PLAYMAINTHRUST,false);
		SetThrusterIsp(th_Jets[0],4*ISP_JET_VAC);
		}
	if (GetThrusterGroupLevel(THGROUP_HOVER) > 0.0)
		SoundOptionOnOff3(MyID,PLAYMAINTHRUST,true);

	if ((ThrusterActive < 0.1) && (GetAltitude() > 24000.0)) {
		SetControlSurfaceLevel(AIRCTRL_ELEVATORTRIM,trimlast);
		SetControlSurfaceLevel(AIRCTRL_ELEVATOR,0.0);
		SetControlSurfaceLevel(AIRCTRL_AILERON,0.0);
		SetControlSurfaceLevel(AIRCTRL_RUDDER,0.0);
	}
	SetThrusterMax0 (th_Jets[0],ActualThrust);

};


void NF104::SetClassCaps(  FILEHANDLE cfg )
{

	CheckCentralMessage();

	SetMainStage();
}

void NF104::SaveState( FILEHANDLE scn )
{
	char cbuf[256];

	SaveDefaultState (scn);
	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* Call this after the normal SaveDefautState() to record attached payloads
	/* it the scenario file */


//	Undock(ALLDOCKS);

	sprintf (cbuf, "%d %0.4f", gear_status, gear_proc);
	oapiWriteScenario_string (scn, "GEAR", cbuf);
	sprintf (cbuf, "%d %0.4f", brakes_status, brakes_proc);
	oapiWriteScenario_string (scn, "BRAKES", cbuf);
	sprintf (cbuf, "%d %0.4f", flaps_status, flaps_proc);
	oapiWriteScenario_string (scn, "FLAPS", cbuf);

	/* save payload lines */
	SaveDefaultStateEx(scn);
}

void NF104::LoadState(  FILEHANDLE scn, void *vs )
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) 
	{
		/* CVE-lite function ! */
		/* IMPORTANT */
		/* Call this for every line in the scenario file. It will automatically
		/* deduce which lines are for it, and load the payloads from the file. */

		/* load payload lines */

		if (!strnicmp (line, "GEAR", 4)) {
		sscanf (line+4, "%d%lf", &gear_status, &gear_proc);
		} else if (!strnicmp (line, "BRAKES", 6)) {
		sscanf (line+6, "%d%lf", &brakes_status, &brakes_proc);
		} else if (!strnicmp (line, "FLAPS", 5)) {
		sscanf (line+5, "%d%lf", &flaps_status, &flaps_proc);

		} else ParsePayloadString(line);

		ParseScenarioLineEx (line, vs);
	}

	/* always set the stage whenever something may have changed */
	SetMainStage();

	// Double-tap all the animations so they display right
	RevertLandingGear();
	RevertLandingGear();

	RevertBrakes();
	RevertBrakes();

	RevertFlaps();
	RevertFlaps();

	if (brakes_proc == 1.0)
		BrakesOn = true;

}


void NF104::PlaceComponents()
{
	COMMON_MESH tmp;

	tmp.MeshHandle(hNF104);
	tmp.MeshPosition(NF104_OFS);
	AddMeshEx (tmp);


}

/* IMPORTANT */
/* Always have a function that you can call as many times as you like that will re-draw
/* all you meshes, reset all your thrusters and check to make sure the propellent is
/* there. This should be called after any configuration change, or any GDPS line (see
/* TimeStep(), above) */

void NF104::SetMainStage( )
{

	if (!AnimationsMade)
		MakeAnimations();

	SetNF104Stage();
}



void NF104::SetNF104Stage( )
{


	/* CVE-lite function ! */
	/* ClearMeshesEx() will get rid of all the normal meshes, plus anything you've defined
	/* with the COMMON_MESH class (not used here) */

	/* erase all meshes and thrusters so we can start afresh */
	ClearMeshesEx();

	PARTICLESTREAMSPEC contrail = 
	{
		0, 0.5, 9, 25.0, 0, 0.5, -.5, 0.5, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 0.1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	PARTICLESTREAMSPEC exhaust_main = {
		0, 3, 7, 350, 0.1, 0.2, 28, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	};
	exhaust_main.tex = oapiRegisterParticleTexture ("Contrail1");

	SetSize (6);

	/* CVE-lite function ! */
	/* IMPORTANT */
	/* double CalculateMass() will return the all-up mass of the combined payloads and
	/* upper stages. Simply add this to your normal dry mass. */
	
	/* always add payload mass [CalculateMass()] to the vessel's empty mass */
	SetEmptyMass ( NF104_DRYMASS + atchmass + CalculateMass() );


	SetTouchdownPoints (_V(0,-1.68,7.2), _V(-3.6,-1.80,-0.66), _V(3.6,-1.80,-0.66));
	SetPMI (_V(20.66,20.86,0.8));
	SetCrossSections (_V(26.06,39.05, 3.03));
	SetCW (0.07, 0.2, 1.0, 1.0);
	SetRotDrag (_V(0.2,0.2,0.15));
	SetPitchMomentScale (1e-6);
	SetBankMomentScale (1e-6);
//	SetLiftCoeffFunc (0); 
	SetTrimScale (0.025);
	SetCameraOffset(CAMERA_OFS);
	MakeAirfoils();
//	SetMeshVisibilityMode(3,MESHVIS_ALWAYS);

	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* Draw() should be called to render all the loaded payloads, plus anything you've
	/* added with the COMMON_MESH class (not used here) */

	/* call draw to render the payloads, plus anything you've added from COMMON_MESH */
	PlaceComponents();
	Draw();
//	SetMeshVisibilityMode(0,MESHVIS_ALWAYS);

	/* we're not using COMMON_MESH today, so draw our own */
//	AddMesh(hNF104, &VECTOR_ZERO );
	

	/* CVE-lite variable ! */
	/* IMPORTANT */
	/* Use the keyword double InitFuelLevel multipled by the maximum mass to have the
	/* fuel levels automatically set from the GPDS payload string */

	/* verify that the propellent hasn't been created before making it */

	if (ph_LH2==NULL)  
		ph_LH2  = CreatePropellantResource(LH2_PROP_MASS,LH2_PROP_MASS,1.0); 
	if (ph_Jets==NULL)  
		ph_Jets  = CreatePropellantResource(JET_PROP_MASS,JET_PROP_MASS,1.0); 
	if (ph_RCS==NULL)  
		ph_RCS  = CreatePropellantResource(RCS_PROP_MASS,RCS_PROP_MASS,1.0); 

	double mainthrust = GetThrusterGroupLevel(THGROUP_MAIN);
	double hoverthrust = GetThrusterGroupLevel(THGROUP_HOVER);

	ClearThrusterDefinitions();
	SetDefaultPropellantResource (ph_Jets); 
	SetMaxWheelbrakeForce (7.5*GetMass());

	th_LH2[0] = CreateThruster ( _V( 0.0,0.838,-7.14), _V(0,-0.11315,0.9935), THRUST_FIRST_VAC, ph_LH2, ISP_FIRST_VAC, ISP_FIRST_SL);
//	AddExhaustStream (th_LH2[0], _V(0.0,0.0,-6.03), &contrail);
//	AddExhaustStream (th_first[0], _V(0.72,0,-7), &exhaust_main);
	AddExhaust (th_LH2[0], 6.0, 0.5, _V( 0.0,0.838,-7.14), _V(0,0.11315,-0.9935), tex);

	thg_LH2 = CreateThrusterGroup (th_LH2, 1, THGROUP_HOVER);

	th_Jets[0] = CreateThruster ( _V( 0.0,0.0,-6.25), _V(0,0,1), JET_THRUST, ph_Jets, ISP_JET_VAC, ISP_JET_VAC);
	AddExhaustStream (th_Jets[0], _V( 0.0,0.0,-6.25), &contrail);
//	AddExhaustStream (th_first[1], _V(-0.72,0,-7), &exhaust_main);

	thg_Jets = CreateThrusterGroup (th_Jets, 1, THGROUP_MAIN);

	SetAttControlsNF104();
	EnableTransponder(true);

	SoundOptionOnOff3(MyID,PLAYMAINTHRUST,false);
	SetThrusterGroupLevel(THGROUP_MAIN,mainthrust);
	SetThrusterGroupLevel(THGROUP_HOVER,hoverthrust);

}




int NF104::ConsumeBufferedKey(DWORD dwKey, bool fPressed, const char *keystate)
{
	if (!fPressed) return 0; // only process keydown events
	if (KEYMOD_SHIFT (keystate)) 
	{
	}
	else if (KEYMOD_CONTROL (keystate)) 
	{
		// insert ctrl key combinations here
		if (KEYDOWN (keystate, OAPI_KEY_E))
		{ 
			if (oapiAcceptDelayedKey (OAPI_KEY_E, 1.0))
				// Eject Eject Eject!!!!

			if (!bEjected) {

				PlayVesselWave3(MyID,PLAYBANG);
				bEjected = true;
				VESSELSTATUS vs;

				GetStatus (vs);
				VECTOR3 rofs, rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
				VECTOR3 ofs = ASTRO_OFS;
				VECTOR3 vel = _V(0,-42.0,0.0);
	//			rofs = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
	//			rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
				Local2Rel (ofs, vs.rpos);
				GlobalRot (vel, rofs);
				vs.rvel.x = rvel.x+rofs.x;
				vs.rvel.y = rvel.y+rofs.y;
				vs.rvel.z = rvel.z+rofs.z;
				vs.vrot.x = 0.0;
				vs.vrot.y = 0.0;
				vs.vrot.z = 0.0;
				char name[256];
				strcpy (name, GetName()); strcat (name, "-Pilot");
				oapiCreateVessel (name, "MOOSE/EJECTSEAT", vs);

//				VESSEL::Create (name, "MOOSE/EJECTSEAT", vs);
				OBJHANDLE hvessel;
				hvessel=oapiGetVesselByName(name);
				if (hvessel != 0) {
					VESSEL *hv;
					hv = oapiGetVesselInterface(hvessel);
					hv->SetThrusterGroupLevel (THGROUP_HOVER,0.1);
					oapiSetFocusObject(hvessel);
					hv->SetFuelMass(0.01);
				}
				SetMainStage();
			}
																
			return 1;
		}
		return 0;
	}
	else {
		if (KEYDOWN (keystate, OAPI_KEY_J)) 
		{ // "Join Stage"
			if (oapiAcceptDelayedKey (OAPI_KEY_J, 1.0))
			{
			/* CVE-lite function ! */
			/* IMPORTANT */
			/* call DeployPayload() to jettison the GPDS payloads in sequence */
	
			AtchV = DeployPayload();
			ignitioncount = 1;

			/* IMPORTANT */
			/* Always re-set the stage after any kind of change */
			SetMainStage();

			return 1;
			}
		}
		if (KEYDOWN (keystate, OAPI_KEY_F)) 
		{ // Toggle the Gear
			if (oapiAcceptDelayedKey (OAPI_KEY_F, 1.0))
			{
				RevertFlaps ();
				StopVesselWave3(MyID,PLAYFLAPS);
				PlayVesselWave3(MyID,PLAYFLAPS);
			}
			return 1;
		}
		if (KEYDOWN (keystate, OAPI_KEY_G)) 
		{ // Toggle the Gear
			if (oapiAcceptDelayedKey (OAPI_KEY_G, 1.0))
			{
				RevertLandingGear ();
				StopVesselWave3(MyID,PLAYGEAR);
				PlayVesselWave3(MyID,PLAYGEAR);
			}
			return 1;
		}
		if (KEYDOWN (keystate, OAPI_KEY_B)) 
		{ // Toggle the brakes
			if (oapiAcceptDelayedKey (OAPI_KEY_B, 1.0))
			{
//				if (BrakesOn)
//					BrakesOn = false;
//				else BrakesOn = true;
				RevertBrakes();
				if (brakes_status == DOOR_OPENING)
					BrakesOn = true;
				else BrakesOn = false;
			}
			return 1;
		}

/*		if (KEYDOWN (keystate, OAPI_KEY_E)) 
		{ 
			if (oapiAcceptDelayedKey (OAPI_KEY_E, 1.0))
				{
				if (hatch_status == DOOR_OPEN)
					{
		//				char name[256];
		//				OBJHANDLE hvessel;
		//				strcpy (name, GetName()); strcat (name, "-EV1");
		//				hvessel=oapiGetVesselByName(name);
		//				if (hvessel == 0)
		//				{
		//					bEVA = true;
							SeparateEVA();
							SetMainStage();
		//					hvessel=oapiGetVesselByName(name);
		//					if (hvessel != 0)
		//						oapiSetFocusObject(hvessel);
		//				}
		//				else
		//				{
		//					oapiDeleteVessel(hvessel);
		//					bEVA = false;
		//					SetMainStage();
		//				};
					}
			}
		return 1;
		}
*/
	}

	return 0;
}




int NF104::ConsumeKey( const char *keystate )
{

	if (KEYMOD_SHIFT (keystate)) 
	{
						
		return 0; // shift combinations are reserved
	}
	else if (KEYMOD_CONTROL (keystate)) 
	{
		return 0;
	}
	else 
	{ // unmodified keys
	
		return 0;

	}
	return 0;
}




void NF104::SetAttControlsNF104()
{
	bAttSet = true;
	THRUSTER_HANDLE th_rot[2];

	th_rot[0] = CreateThruster (_V(0,-0.97, 5.25), _V(0, 1,0), MAX_ATT_FIRST, ph_RCS, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,-0.97,-5.25), _V(0,-1,0), MAX_ATT_FIRST, ph_RCS, ISP_FIRST_SL);
	thg_pitchup=CreateThrusterGroup (th_rot, 2, THGROUP_ATT_PITCHUP);

	th_rot[0] = CreateThruster (_V(0,0, 5), _V(0,-1,0), MAX_ATT_FIRST, ph_RCS, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,0,-5), _V(0, 1,0), MAX_ATT_FIRST, ph_RCS, ISP_FIRST_SL);
	thg_pitchdown=CreateThrusterGroup (th_rot, 2, THGROUP_ATT_PITCHDOWN);

	th_rot[0] = CreateThruster (_V(0,0, 4), _V(-1,0,0), MAX_ATT_FIRST, ph_RCS, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,0,-4), _V( 1,0,0), MAX_ATT_FIRST, ph_RCS, ISP_FIRST_SL);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_YAWLEFT);

	th_rot[0] = CreateThruster (_V(0,0, 4), _V( 1,0,0), MAX_ATT_FIRST, ph_RCS, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(0,0,-4), _V(-1,0,0), MAX_ATT_FIRST, ph_RCS, ISP_FIRST_SL);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_YAWRIGHT);

	th_rot[0] = CreateThruster (_V(13,0,0), _V(0,-1,0), MAX_ATT_FIRST/4, ph_RCS, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(-13,0,0), _V(0, 1,0), MAX_ATT_FIRST/4, ph_RCS, ISP_FIRST_SL);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_BANKRIGHT);

	th_rot[0] = CreateThruster (_V(13,0, 0), _V(0, 1,0), MAX_ATT_FIRST/4, ph_RCS, ISP_FIRST_SL);
	th_rot[1] = CreateThruster (_V(-13,0,0), _V(0, -1,0), MAX_ATT_FIRST/4, ph_RCS, ISP_FIRST_SL);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_BANKLEFT);


}

