/**
*
* $Id: CVEL-NF104.h,v 0.1 2004/11/20 18:03:45 erv Exp $
* (c) 2004 Sputnik -- erik_h_anderson@hotmail.com
* Based on CVEL Centaur-G code
* Ervins Reinverts -- ervins@gmail.com
* Based on CVEL Fregat code
* (c) 2004 David Henderson -- orbiter@aibs.org.uk
* Licence LGPL -- http://www.gnu.org/licenses/lgpl.txt
*
* You may use this source code in your addon provided that if you
* publish your addon either privately or publicly the end user
* recieves a copy of the source code including any modifications
* you may have made. 
*
*/


#include "../../include/cvel.h"
#include "../../include/cvelmessage.h"
#include "../../include/orbitersdk.h"
#include "../../include/OrbiterSoundSDK3.h"
#include <cstdio>
#include <cmath>
#include <string>

#define InitFuelLevel 1

// All data from astronautix.com and http://www.bristolspaceplanes.com/projects/ascender.shtml
const double ISP_FIRST_SL		=     230.0*G; 
const double ISP_FIRST_VAC		=     240.0*G; //***
const double ISP_RCS_VAC		=      55.0*G; //***
const double ISP_JET_VAC		=    6500.0*G;
const double THRUST_FIRST_VAC	=   31500.0;
const double MAX_ATT_FIRST		=      32.0*G;
const double JET_THRUST			=    9115.0*G*1.5;
const double LH2_PROP_MASS		=     530.0;
const double JET_PROP_MASS		=    2721.0;
const double NF104_DRYMASS		=    7483.0;	// Total gross mass 4500 kg 
const double RCS_PROP_MASS		=      50.0; //***

const VECTOR3 NF104_OFS		= _V(0,0,0);
const VECTOR3 CAMERA_OFS	= _V( 0, 0.55,6.25);
const VECTOR3 ASTRO_OFS		= _V( 0,-0.10,6.2);

const double GEAR_OPERATING_SPEED = 0.25;
const double FLAPS_OPERATING_SPEED = 0.25;
const double BRAKES_OPERATING_SPEED = 0.5;

#define PLAYGEAR					 20		// the gear sound
#define PLAYJET						 26     // the jet noise sound
#define PLAYBANG					 27     // the jettison sound
#define PLAYFLAPS					 28     // the flaps sound

/* make your own vessel class (named NF104 here, for example)
   subclassed from COMMON_VESSEL_EX */

class NF104 : public COMMON_VESSEL_EX {
public:
	NF104 (OBJHANDLE hObj, int fmodel);
	~NF104();

	enum DoorStatus { DOOR_CLOSED, DOOR_OPEN, DOOR_CLOSING, DOOR_OPENING } gear_status, brakes_status, flaps_status;

	void MakeAnimations();
	void MakeAirfoils();
	void MakeControlSurfaces();
//	void BoosterSwitchToJets();
	void UpdateAnimations();
	void ActivateLandingGear (DoorStatus action);
	void RevertLandingGear();
	void ActivateBrakes (DoorStatus action);
	void RevertBrakes();
	void ActivateFlaps (DoorStatus action);
	void RevertFlaps();
	void SetGearParameters (double state);
	void TimeStep(  double simt ); // need a timestep loop to conduct the GPDS polling
	void AirBreathingEngine();
	void VisualCreated( VESSEL *vessel, VISHANDLE vis, int refcount ); // must use VisualCreated() instead of ovcVisualCreated()
	void SetClassCaps(  FILEHANDLE cfg );
	void LoadState(  FILEHANDLE scn, void *vs );
	void SaveState(  FILEHANDLE scn );
	void ProgramAuto();
	void PlaceComponents();

	/* CVE-lite requirement ! */
	// Create a function that can be called as many times as you like. It is
	// vital to have a function like this that cleans out all the
	// thrusters and meshes and draws/recreates them again.
	// This is so that you can respond immediately to GPDS requests.
	// Here I've called it SetMainStage() it can be Refresh() or anything you like.
	void SetMainStage(); 
	void SetNF104Stage();  
	void SetAttControlsNF104();    // this is to set up the thrusters. I like to keep them separate.
	bool AzInput (void *id, char *str, void *data);
	bool IncInput (char *rstr);
	int  ConsumeKey(  const char *keystate );
	int  ConsumeBufferedKey(DWORD dwKey, bool fPressed, const char *keystate);
	double pitchinput, yawinput, rollinput;
	VESSELSTATUS vst;


// Overloaded callback functions
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkLoadStateEx (FILEHANDLE scn, void *vs);
	void clbkSaveState (FILEHANDLE scn);
	void clbkPostStep (double simt, double simdt, double mjd);
	int  clbkConsumeBufferedKey (DWORD key, bool down, char *kstate);
	void clbkVisualCreated (VISHANDLE vis, int refcount);
	void clbkVisualDestroyed (VISHANDLE vis, int refcount);
//	void clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, HDC hDC);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);

protected:



	PROPELLANT_HANDLE ph_LH2, ph_Jets, ph_RCS;
	THRUSTER_HANDLE   th_LH2[1], th_Jets[2];
	THGROUP_HANDLE    thg_LH2, thg_Jets,thg_pitchup, thg_pitchdown;
	MESHHANDLE		  hNF104;
	ANIMATIONCOMPONENT_HANDLE Noz1, Noz2;
//	ANIMATIONCOMPONENT_HANDLE wing1, wing2, wing3, wing4, Gear1, Gear2, Gear3, Gear4, Gear5, Gear6, Gear7,
//		PLDr1, PLDr2, Hatch1, Jets1, Nozl1;
//	ANIMATIONCOMPONENT_HANDLE wing11, wing12, wing13, wing14, Gear11, Gear12, Gear13, Gear14, Gear15, Gear16, Gear17,
//		Gear18, Gear19, Jets11, Jets12, Jets13, Nozl11, Elev11, Rudd11, Elev12, Rudd12, Elev13, Rudd13, Elev14, Rudd14;
	SURFHANDLE		  tex,texRCS, retex;
	UINT anim_rudder,anim_laileron,anim_raileron,anim_elevator,anim_elevatortrim, anim_flaps;
	UINT anim_gear;     // handle for landing gear animation
	UINT anim_pldoors;     // handle for payload door animation
	UINT anim_brakes;	// handle for speedbrake animation
	UINT anim_nozzles;	// handle for nozzles animation
	UINT anim_trolley, anim_trolley2;	// handle for trolley animation
	double gear_proc;
	double brakes_proc, flaps_proc;
	bool BrakesOn, havesound, 
		bAbort, AbortChecked, bAttSet,bBoosterMade, bAutoProgrammed, bControlsMade;
	UINT Exh1,Exh2,Exh3,Exh4;
	PSTREAM_HANDLE stream_a_1,stream_b_1,stream_a_2,stream_b_2,stream_a_3,stream_b_3,stream_a_4,stream_b_4;
	VECTOR3 Vold, Vnew;
	double dt, oldsimt;
	int MyID;
	ATTACHMENTHANDLE Baypos, sat_attach, Nose_atch1, Nose_atch2;
	VESSEL *AtchV;
	DWORD i, j;
	double atchmass;
	DOCKHANDLE hDock, Dock0, Dock1, Dock2;
	bool Firstsweep;
	double tempP, tempY, tempR;
	VECTOR3 AngularVel;
	bool autopilot;
	bool engineignited, bgoingsouth;
	double last_simt;
	double ignition_time;
	double density;
	double trimlast;
	bool bEjected;
	double displaytime, simt, G_level, d_heading;
	bool AnimationsMade;
	int ignitioncount;
};








/* basic vessel class setup */

void NF104::clbkSetClassCaps (FILEHANDLE cfg)
{
	SetClassCaps(cfg);
}

void NF104::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
	LoadState(  scn, vs );
}

void NF104::clbkSaveState (FILEHANDLE scn)
{
	SaveState(  scn );
}

void NF104::clbkPostStep (double simt, double simdt, double mjd)
{
	TimeStep(simt);
}

void NF104::clbkVisualCreated (VISHANDLE _vis, int refcount)
{
	SetMainStage( );
}

void NF104::clbkVisualDestroyed (VISHANDLE _vis, int refcount)
{
	SetMainStage( );
}

int NF104::clbkConsumeBufferedKey (DWORD dwKey, bool fPressed, char *keystate)
{
	int ret = ConsumeBufferedKey( dwKey, fPressed, keystate );
	return ret;
}


/**
 * CHANGES BY URWUMPE:
 * Display equivalent airspeed in knots, pressure height and radar height (below 2500 ft), both in ft
 * Flash brakes indicator while in transition
 */
bool NF104::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	// show gear deployment status
	if (gear_status == DOOR_OPEN || (gear_status >= DOOR_CLOSING && fmod (oapiGetSimTime(), 1.0) < 0.5)) {
		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;
		if (cx >= -d*3 && cx < hps->W+d*3 && cy >= d && cy < hps->H+d*5) {
			skp->Rectangle (cx-d/2, cy-d*5, cx+d/2, cy-d*4);
			skp->Rectangle (cx-d*3, cy-d*2, cx-d*2, cy-d);
			skp->Rectangle (cx+d*2, cy-d*2, cx+d*3, cy-d);
		}
	}

	// show brakes deployment status
	if (BrakesOn || brakes_proc > 0) {
		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;
		if((((!BrakesOn && brakes_proc > 0) || (BrakesOn && brakes_proc < 1.0)) && 
				fmod (oapiGetSimTime(), 0.5) < 0.25)||(BrakesOn && brakes_proc == 1.0))
		{
			//Flash indicator
			if (cx >= -d*3 && cx < hps->W+d*3 && cy >= d && cy < hps->H+d*5) {
				skp->Text (cx-d*9, cy+d*1.2, "BRAKES", 6);
			}
		}
	}
	//Draw Equivalent airspeed & pressure altitude
	if(GetAtmPressure() > 0.0)
	{
		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;
		char pszBuffer[40];
		int bx = 0.5 * cx;
		const double EAS_EXP = 2.0/7.0;
		const double P0 = 1.01325E6;

		double fEAS = 661.48 * sqrt(5.0 * GetAtmPressure()/ATMP * (pow(GetDynPressure() / GetAtmPressure() + 1, EAS_EXP) - 1.0));

		sprintf(pszBuffer, "%6.1f KEAS", fEAS);

		skp->Text(bx, cy - d, pszBuffer, 11);
	}
	if (GetAtmPressure() > 7000.0)				  // About 60,000'; use pressure altitude
	{
		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;
		char pszBuffer[40];
		int bx = 0.5 * cx;
		const double EAS_EXP = 2.0/7.0;
		const double P0 = 1.01325E6;

		bx = 0.5 * (cx + hps->W) - d * 8;

		double fHp = (1.0 - pow(GetAtmPressure()/ATMP, 0.190263)) * (288.15/0.00198122);

		long H_press = (long)(fHp);
		H_press = H_press - H_press % 50;

		sprintf(pszBuffer, "%6d'  ", H_press);

		skp->Text(bx, cy - d, pszBuffer, 9);
	} else {									// Do inertial altitude instead
		int d = hps->Markersize/2;		
		int cx = hps->CX, cy = hps->CY;
		char pszBuffer[40];
		int bx = 0.5 * cx;
		const double EAS_EXP = 2.0/7.0;
		const double P0 = 1.01325E6;

		bx = 0.5 * (cx + hps->W) - d * 8;

		double fIalt = GetAltitude()/0.3048;

		long I_Alt = (long)(fIalt);
		I_Alt = I_Alt - I_Alt % 50;

		sprintf(pszBuffer, "%6d' I", I_Alt);

		skp->Text(bx, cy - d, pszBuffer, 9);
	}

	//Draw Radar altitude
	if(GetAltitude() < 2500.0 * 0.3048)
	{
		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;
		char pszBuffer[40];
		int bx = 0.5 * (cx + hps->W) - d * 8;
		const double GEAR_Y = -1.35;

		sprintf(pszBuffer, "%6dR", static_cast<long>((GetAltitude() + GEAR_Y)/0.3048));

		skp->Text(bx, cy + d, pszBuffer, 7);
	}

	if(GetThrusterLevel(th_LH2)> 0.0 && GetPropellantFlowrate(ph_LH2) > 0.0 && GetPropellantMass(ph_LH2) > 0.0)
	{
		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;
		char pszBuffer[40];
		int bx = 0.5 * (cx + hps->W);

		double t_co = GetPropellantMass(ph_LH2)/GetPropellantFlowrate(ph_LH2);

		sprintf(pszBuffer, "CO %02d:%02d", static_cast<long>(t_co/60.0), static_cast<long>(fmod(t_co, 60.0)));

		skp->Text(bx, 80.0, pszBuffer, strlen(pszBuffer));
		
	}

	//Draw G-Meter

		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;

		char pszBuffer[40];
		int bx = 0.5 * cx;

		VECTOR3 ForceVec, WeightVec;
		GetForceVector(ForceVec);
		GetWeightVector(WeightVec);
		ForceVec -= WeightVec;
		G_level = sqrt(pow(ForceVec.x,2)+pow(ForceVec.y,2)+pow(ForceVec.z,2)) / (9.81*GetMass());

		sprintf(pszBuffer, "  %.2f G", G_level);

		skp->Text(bx, cy + d, pszBuffer, strlen(pszBuffer));


	// show flaps deployment status
	if (flaps_status == DOOR_OPEN || (flaps_status >= DOOR_CLOSING && fmod (oapiGetSimTime(), 1.0) < 0.5)) {
		int d = hps->Markersize/2;
		int cx = hps->CX, cy = hps->CY;
		if (cx >= -d*3 && cx < hps->W+d*3 && cy >= d && cy < hps->H+d*5) {
			skp->Text(cx+d*5, cy+d*1.2, "FLAPS", 5);
		}
	}


	return true;
}

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new NF104 (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (NF104*)vessel;
}

